#include <iostream>
#include <string>
#include <sstream>
#include <boost/asio.hpp>
#include <cctype>
#include <cstdio>
#include "basic_nngs.h"

#include "go_board.h"
#include "search.h"
#include "make_pattern.h"
#include "gtp.h"

BasicNNGS::BasicNNGS()
{
    ai_name_ = "randomz";
    ai_color_ = 0; // default black
    initializeBasicNNGS();
    turn_count_ = 0;
    search_.clearNodes();
}

void BasicNNGS::initializeBasicNNGS()
{
    for (int i = 0; i < kboard_square_size; ++i) {
        std::stringstream ss;
        int pos = board9x9_index[i];
        int x = (pos % kexpanded_board_size) - 2;
        if (x >= 8) {
            ++x;
        }

        int y = pos / kexpanded_board_size - 2;
        ss << (char)('A' + x) << (kboard_size - y);
        gtp_to_pos_[ss.str()] = pos;
        pos_to_gtp_[pos] = ss.str();

        // std::cerr << pos << " " << ss.str() << std::endl;
    }
    pos_to_gtp_[0] = "pass";
}

void BasicNNGS::loop()
{
    std::cin.clear();
    std::cout.flush();

    int counts = 0;
    std::string color_str = (ai_color_ == 0 ? "B" : "W");
    // boost::asio::ip::tcp::iostream stream(("jsb.cs.uec.ac.jp", "9696");
    boost::asio::ip::tcp::iostream stream("192.168.11.213", "9696");
    // boost::asio::ip::tcp::iostream stream(("localhost", "9696");
    std::string opponent_name;

    if (!stream) {
        std::cerr << "could not connect to nngs server" << std::endl;
        return;
    }

    std::cerr << "please enter name" << std::endl;
    std::cin >> ai_name_;
    if (ai_name_.find("_") != std::string::npos) {
        std::cerr << "can not be regnized as name" << std::endl;
        return;
    }

    stream << ai_name_ << "\n\n" << std::flush;
    stream << "set client TRUE" << std::endl << std::endl << std::flush;
    std::cerr << "please enter my color B or W" << std::endl;
    std::cin >> color_str;
    ai_color_ = ((color_str == "B" || color_str == "b") ? 0 : 1);
    std::cerr << "please enter opponent's name" << std::endl;
    std::cin >> opponent_name;
    stream << "match " << opponent_name << " " << (ai_color_ == Black ? "B" : "W") << " 9 0 1" << std::endl << std::endl << std::flush;
    std::cerr << "match " << opponent_name << " " << (ai_color_ == Black ? "B" : "W") << " 9 0 1" << std::endl;

    if (ai_color_ == Black) {
        counts = 2;
    }

    // Intersection ai_color = color_str == "B" ? BLACK : WHITE;
    bool pass_flag = false;

    while(true) {
        bool is_opponen_down = false;
        bool sw = true;
        while(true) {
            std::string mes;
            std::getline(stream, mes);
            std::cerr << mes << std::endl;

            if (mes.size() > 1 && mes[0] == '1' && mes[1] == '5') {
                counts++;
            }

            if (mes.find("'done'") != std::string::npos) {
                is_opponen_down = true;
                break;
            }
            else if (mes.find("No user named") != std::string::npos) {
                std::cerr << "not found " << opponent_name << std::endl;
                is_opponen_down = true;
                break;
            }
            else if (mes.find("resigned") != std::string::npos) {
                std::cerr << "resigned " << std::endl;
                is_opponen_down = true;
                break;
            }

            if (mes.size() < 17 && ((counts == 2 && sw) || counts == 4)) {
                sw = false;

                if (mes.find("(B)") != std::string::npos) {
                    std::cout << "play B " << mes.substr(11) << std::endl << std::endl;
                    int color = Black;

                    if (mes.size() > 11 && (mes[12] == 'a' || mes[12] == 'A')) {
                        // pass
                        go_board_.setStone(0, color);
                        pass_flag = true;
                        go_board_.initializeKoPoint();
                    }
                    else {
                        pass_flag = false;
                        mes.resize(mes.size() - 1);

                        int move = gtp_to_pos_[mes.substr(11)];

                        // cerr << "move after : " << move << endl;

                        go_board_.setStone(move, color);
                        ++turn_count_;
                        // go_board_.printGoBoard();
                    }
                }
                else if (mes.find("(W)") != std::string::npos) {
                    std::cout << "play W " << mes.substr(11) << std::endl << std::endl;
                    int color = White;
                    if (mes.size() > 11 && (mes[12] == 'a' || mes[12] == 'A')) {
                        go_board_.setStone(0, color);   // color
                        ++turn_count_;
                        pass_flag = true;
                        go_board_.initializeKoPoint();
                    }
                    else {
                        pass_flag = false;
                        mes.resize(mes.size() - 1);
                        int move = gtp_to_pos_[mes.substr(11)];

                        go_board_.setStone(move, color);
                        // go_board_.printGoBoard();
                    }
                }
            }


            // iregural move
            if (mes.size() > 0 && mes[0] == '5') {
                counts += 2;
            }


            std::cerr << "counts " << counts << std::endl;

            if (counts == 4) {
                break;
            }

        }
        if (is_opponen_down) {
            break;
        }
        counts = 0;

        std::cerr << "input cmd" << std::endl;
        // getline(cin, mes);
        // s << mes << "\n\n" << flush;

        search_.setGoBoard(go_board_);
        // int move = search_.doRandomSearch(ai_color_);
        int move = search_.doCEDECSearch(ai_color_, turn_count_);

        std::cerr << "move : " << move << std::endl;

        if (move == -1) {
            std::cerr << "= resign" << std::endl;
            stream << "resign" << std::endl << std::endl << std::flush;
            break;
        }


        std::string set_pos =  pos_to_gtp_[move];
        ++turn_count_;

        std::cerr << move << " " << set_pos << std::endl;
        stream << set_pos << std::endl << std::endl << std::flush;

        search_.clearNodes();

        if (set_pos == "pass" && pass_flag) {
            break;
        }
    }

    stream << "done" << "\n\n" << std::flush;
    std::cerr << "fine" << std::endl;
    // go_board_.printResult();
    sleep(2);
}
