#include <iostream>
#include <array>
#include <vector>
#include <string>
#include <fstream>
#include <algorithm>
#include <iomanip>
#include <bitset>
#include <unordered_map>
#include "go_board.h"
#include "search.h"
#include "global.h"

extern MakePattern make_pattern;

GoBoard::GoBoard()
{
    initializeGoBoard();
}

void GoBoard::initializeGoBoard()
{
    clearGoBoard();
    clearGoStrings();
    ko_position_ = 0;
    pre_position_ = 0;
    second_pre_position_ = 0;
    thrid_pre_position_ = 0;
    stone_count_ = 0;
}

void GoBoard::printGoBoard() const
{
    for (int i = 0; i <  karray_board_size; ++i) {
        if (!(i % kexpanded_board_size)) {
            std::cerr << std::endl;
        }
        std::cerr << " " << (go_board_[i] == Black ? "B" : go_board_[i] ==White ? "W" : go_board_[i] == Empty ? "+" : "N");
    }
    std::cerr << std::endl << std::endl;
}

void GoBoard::setStone(int pos, int color)
{
    if (pos == 0) {
        thrid_pre_position_ = second_pre_position_;
        second_pre_position_ = pre_position_;
        pre_position_ = pos;

        return;
    }
    if (go_board_[pos] != Empty) {
        return;
    }

    thrid_pre_position_ = second_pre_position_;
    second_pre_position_ = pre_position_;
    pre_position_ = pos;
    if (pos == 0) {
        // pass
        ko_position_ = 0;
        return;
    }

    ++stone_count_;
    go_board_[pos] = color;

    int connection[4] = {0};
    int inverse_connection[4] = {0};
    int parent = 0;
    int liberty_count = 0;
    int inverse_color = (color == Black ? White : Black);
    int parent_index = 0;
    int remove_stone_count = 0;
    int remove_stone_pos = 0;
    int remove_around_count = 0;
    int tmp_pos = 0;
    int tmp_color = 0;
    int around_pos[4] = {getEast(pos), getWest(pos), getNorth(pos), getSouth(pos)};

    if ((tmp_color = go_board_[tmp_pos = around_pos[0]]) == Empty) {
        ++liberty_count;
    }
    else if (tmp_color == color) {
            parent = connection[0] = go_string_board_[tmp_pos];
            parent_index = 0;
    }
    else if (tmp_color == inverse_color) {
            inverse_connection[0] = go_string_board_[tmp_pos];
    }


    if ((tmp_color = go_board_[tmp_pos = around_pos[1]]) == Empty) {
        ++liberty_count;
    }
    else if (tmp_color == color) {
            parent = connection[1] = go_string_board_[tmp_pos];
            parent_index = 1;
    }
    else if (tmp_color == inverse_color) {
            inverse_connection[1] = go_string_board_[tmp_pos];
    }


    if ((tmp_color = go_board_[tmp_pos = around_pos[2]]) == Empty) {
        ++liberty_count;
    }
    else if (tmp_color == color) {
            parent = connection[2] = go_string_board_[tmp_pos];
            parent_index = 2;
    }
    else if (tmp_color == inverse_color) {
            inverse_connection[2] = go_string_board_[tmp_pos];
    }

    if ((tmp_color = go_board_[tmp_pos = around_pos[3]]) == Empty) {
        ++liberty_count;
    }
    else if (tmp_color == color) {
            parent = connection[3] = go_string_board_[tmp_pos];
            parent_index = 3;
    }
    else if (tmp_color == inverse_color) {
            inverse_connection[3] = go_string_board_[tmp_pos];
    }

    // no connection
    if (parent == 0) {
        if (color == Black) {
            go_string_board_[pos] = go_string_index_[0];
            go_string_liberty_count_[go_string_index_[0]] = liberty_count;
            go_string_index_[0] += 2;
            if (go_string_index_[0] >= go_string_liberty_count_.size()) {
                go_string_liberty_count_.resize(go_string_liberty_count_.size() + 100);
            }

        }
        else if (color == White) {
            go_string_board_[pos] = go_string_index_[1];
            go_string_liberty_count_[go_string_index_[1]] = liberty_count;
            go_string_index_[1] += 2;
            if (go_string_index_[1] >= go_string_liberty_count_.size()) {
                go_string_liberty_count_.resize(go_string_liberty_count_.size() + 100);
            }
        }
    }
    else{
        go_string_board_[pos] = parent;
        go_string_liberty_count_[parent] += liberty_count - 1;

        for (int i = 0 ;i < parent_index; ++i) {
            if (connection[i] != 0) {
                if (connection[i] != parent) {
                    std::replace(go_string_board_.begin(), go_string_board_.end(), connection[i], parent);
                    go_string_liberty_count_[parent] += go_string_liberty_count_[connection[i]] - 1;
                    if (connection[0] > 0) {
                        connection[0] = go_string_board_[around_pos[0]];
                    }
                    if (connection[1] > 0) {
                        connection[1] = go_string_board_[around_pos[1]];
                    }
                    if (connection[2] > 0) {
                        connection[2] = go_string_board_[around_pos[2]];
                    }
                    if (go_board_[getSouth(pos)] == color) {
                        connection[3] = go_string_board_[around_pos[3]];
                    }
                }
                else {
                    --go_string_liberty_count_[parent];
                }
            }
        }
    }


    bool is_death = false;
    for (int i = 0; i < 4; ++i) {
        --go_string_liberty_count_[inverse_connection[i]];
        if (go_string_liberty_count_[inverse_connection[i]] == 0) {
            // strings death
            is_death = true;
            for (int j = kexpanded_board_size; j <  karray_board_size - kexpanded_board_size; ++j) {
                if (go_string_board_[j] == inverse_connection[i]) {
                    go_string_board_[j] = 0;
                    go_board_[j] = Empty;
                    --stone_count_;
                    ++remove_stone_count;
                    remove_stone_pos = j;

                    if (go_board_[getEast(j)] == color) {
                        ++go_string_liberty_count_[go_string_board_[getEast(j)]];
                    }
                    if (go_board_[getWest(j)] == color) {
                        ++go_string_liberty_count_[go_string_board_[getWest(j)]];
                    }
                    if (go_board_[getNorth(j)] == color) {
                        ++go_string_liberty_count_[go_string_board_[getNorth(j)]];
                    }
                    if (go_board_[getSouth(j)] == color) {
                        ++go_string_liberty_count_[go_string_board_[getSouth(j)]];
                    }

                }
            }
        }
    }

    if (remove_stone_count == 1) {
        if (go_board_[around_pos[0]] == inverse_color || go_board_[around_pos[0]] == Wall) {
            ++remove_around_count;
        }
        if (go_board_[around_pos[1]] == inverse_color || go_board_[around_pos[1]] == Wall) {
            ++remove_around_count;
        }
        if (go_board_[around_pos[2]] == inverse_color || go_board_[around_pos[2]] == Wall) {
            ++remove_around_count;
        }
        if (go_board_[around_pos[3]] == inverse_color || go_board_[around_pos[3]] == Wall) {
            ++remove_around_count;
        }
    }
    if (remove_around_count == 3) {
        ko_position_ = remove_stone_pos;
        // std::cerr << "ko ! " << ko_position_ << std::endl;
    }
    else {
        ko_position_ = 0;
    }
}

void GoBoard::printGoStrings()
{
    for (int i = 0; i <  karray_board_size; ++i) {
        if (!(i % kexpanded_board_size)) {
            std::cerr << std::endl;
        }
        std::cerr << " " << go_string_board_[i];
    }

    std::cerr << std::endl << "strings of black" << std::endl;
    for (int i = 1; i < go_string_index_[0]; i += 2) {
        std::cerr << "index : " << i << ", " << go_string_liberty_count_[i] << std::endl;
    }
    std::cerr << "strings of which" << std::endl;
    for (int i = 2; i < go_string_index_[1]; i += 2) {
        std::cerr << "index : " << i << ", " << go_string_liberty_count_[i] << std::endl;
    }

}

bool GoBoard::isLegalMove(int pos, int color) const
{
    if (pos == 0) {
        return true;
    }
    else if (go_board_[pos] != Empty) {
        return false;
    }
    else if (pos == ko_position_) {
        return false;
    }

    // std::unordered_map<int, int> go_string_count;
    std::array<int, 4> go_string_count;
    // std::vector<int> go_string_count(4);
//    go_string_count.reserve(4);
    int around_pos[4] = {getEast(pos), getWest(pos), getNorth(pos), getSouth(pos)};

    if (go_board_[(around_pos[0] = getEast(pos))] == Empty) {
        return true;
    }
    else if (go_board_[(around_pos[1] = getWest(pos))] == Empty) {
        return true;
    }
    else if (go_board_[(around_pos[2] = getNorth(pos))] == Empty) {
        return true;
    }
    else if (go_board_[(around_pos[3] = getSouth(pos))] == Empty) {
        return true;
    }

    go_string_count[0] = go_string_board_[around_pos[0]];
    go_string_count[1] = go_string_board_[around_pos[1]];
    go_string_count[2] = go_string_board_[around_pos[2]];
    go_string_count[3] = go_string_board_[around_pos[3]];


    int inverse_color = (color == Black ? White : Black);
    for (int i = 0; i < 4; ++i) {
        if (go_board_[around_pos[i]] == color) {

            if (go_string_liberty_count_[go_string_count[i]] - std::count(go_string_count.begin(), go_string_count.end(), go_string_count[i])) {
                return true;
            }
        }
        else if (go_board_[around_pos[i]] == inverse_color) {
            if ((go_string_liberty_count_[go_string_count[i]] - std::count(go_string_count.begin(), go_string_count.end(), go_string_count[i])) == 0) {
                return true;
            }
        }
    }
    return false;
}

void GoBoard::readSGF(std::string file_name)
{
    // assume that black is first
    std::ifstream fin(file_name.c_str());

    if (!fin) {
        std::cerr << "not found file" << std::endl;
        return;
    }

    int color = Black;
    char ch, x, y;
    int num_x, num_y;

    while (fin) {
        fin >> ch;
        if (ch == ';') {
            fin >> ch;
            if (ch == 'B' || ch == 'W') {
                fin >> ch;
                if (ch == '[') {
                    fin >> x;
                    if (x != ']') {
                        fin >> y>> ch;
                        if (ch == ']') {
                            num_x = x - 'a' + 2;
                            num_y = y - 'a' + 2;
                            num_y *= kexpanded_board_size;
                            setStone(num_x + num_y, color);
                        }
                    }
                    color = (color == Black ? White : Black);
                }
            }
        }
    }
    fin.close();
}

bool GoBoard::isEye(int pos, int color)
{
    if (pos == 0 || go_board_[pos] != Empty) {
        return false;
    }

    int around_pos[4] = {getEast(pos), getWest(pos), getNorth(pos), getSouth(pos)};
    int point_color;

    if (((point_color = go_board_[around_pos[0]]) == color || point_color == Wall)
            && ((point_color = go_board_[around_pos[1]]) == color || point_color == Wall)
            && ((point_color = go_board_[around_pos[2]]) == color || point_color == Wall)
            && ((point_color = go_board_[around_pos[3]]) == color || point_color == Wall)) {

            if (getLibertyCount(around_pos[0]) > 1 && getLibertyCount(around_pos[1]) > 1
                    && getLibertyCount(around_pos[2]) > 1 && getLibertyCount(around_pos[3]) > 1) {
                return true;
            }
    }


    return false;
}

int GoBoard::getResultChineseRule(int color)
{
    std::array<int, 2> score;
    score.fill(0);
    score[1] = kkomi;
    for (int i = kexpanded_board_size; i <  karray_board_size - kexpanded_board_size; ++i) {
        if (go_board_[i] == Wall) {
            continue;
        }
        else if (go_board_[i] == Black) {
            ++score[0];
        }
        else if (go_board_[i] == White) {
            ++score[1];
        }
        else if (go_board_[i] == Empty) {
            std::array<int, 2> around;
            around.fill(0);
            if (go_board_[getEast(i)] == Black) {
                ++around[0];
            }
            else if (go_board_[getEast(i)] == White) {
                ++around[1];
            }
            if (go_board_[getWest(i)] == Black) {
                ++around[0];
            }
            else if (go_board_[getWest(i)] == White) {
                ++around[1];
            }
            if (go_board_[getNorth(i)] == Black) {
                ++around[0];
            }
            else if (go_board_[getNorth(i)] == White) {
                ++around[1];
            }
            if (go_board_[getSouth(i)] == Black) {
                ++around[0];
            }
            else if (go_board_[getSouth(i)] == White) {
                ++around[1];
            }

            if (around[0] > 0 && around[1] > 0) {
                continue;
            }
            else if (around[0] > 0) {
                ++score[0];
            }
            else if (around[1] > 0) {
                ++score[1];
            }
        }
    }

    if (color == Black) {
        return (score[0] > score[1] ? 1 : 0);
    }
    else if (color == White) {
        return (score[1] >= score[0] ? 1 : 0);
    }
}

double GoBoard::getResultScoreChineseRule()
{
    double score = -kreal_komi;

    for (int i = kboard_start_index; i <  kboard_end_index; ++i) {
        if (go_board_[i] == Black) {
            ++score;
        }
        else if (go_board_[i] == White) {
            --score;
        }
        else if (go_board_[i] == Empty) {
            std::array<int, 2> around = {0, 0};
            ++around[go_board_[getEast(i)]];
            ++around[go_board_[getWest(i)]];
            ++around[go_board_[getNorth(i)]];
            ++around[go_board_[getSouth(i)]];

            if (around[0] && around[1]) {
                continue;
            }
            else if (around[0]) {
                ++score;
            }
            else if (around[1]) {
                --score;
            }
        }
        else if (go_board_[i] == Wall) {
            continue;
        }
    }

    return score;

}

void GoBoard::printPatternGoBoard() const
{
    for (int i = 0; i < karray_board_size; ++i) {
        if (go_board_[i] != Wall) {
            printPattern3x3GoBoard(i);
        }
    }
    std::cerr << std::endl;
}

void GoBoard::printPattern3x3GoBoard(int pos) const
{
    int pattern = pattern_go_board_[pos];
    for (int i = 0; i < 9; ++i) {
        if (!(i % 3)) {
            std::cerr << std::endl;
        }
        int pos = (pattern & 3);
        if (pos == Black) {
            std::cerr << "B";
        }
        else if (pos == White) {
            std::cerr << "W";
        }
        else if (pos == Empty) {
            std::cerr << "+";
        }
        else {
            std::cerr << "N";
        }

        pattern >>= 2;
    }
    std::cerr << pattern_go_board_[pos] << std::endl;
}

void GoBoard::setPatternGoBoard()
{
    for (int i = 0; i < karray_board_size; ++i) {
        if (i / kexpanded_board_size == 0 || i / kexpanded_board_size == 12
                || i % kexpanded_board_size == 0 || i % kexpanded_board_size == 12)  {
            pattern_go_board_[i] = 262143;
            continue;
        }

        int pattern[9] = {go_board_[getEast(getSouth(i))], go_board_[getSouth(i)], go_board_[getWest(getSouth(i))],
                          go_board_[getEast(i)], go_board_[i], go_board_[getWest(i)],
                          go_board_[getEast(getNorth(i))], go_board_[getNorth(i)], go_board_[getWest(getNorth(i))]};
        pattern_go_board_[i] = 0;
        for (int j = 0; j < 9; ++j) {
            pattern_go_board_[i] <<= 2;
            if (pattern[j] == White) {
                ++pattern_go_board_[i];
                // pattern_go_board_[i][1] = 0;
            }
            else if (pattern[j] == Empty) {
                pattern_go_board_[i] += 2;
            }
            else if (pattern[j] == Wall) {
                pattern_go_board_[i] += 3;
            }
        }
    }
}

void GoBoard::setPointPatternGoBoard(int pos)
{
    if (go_board_[pos] == Wall) {
        return;
    }
    int pattern[9] = {go_board_[getEast(getSouth(pos))], go_board_[getSouth(pos)], go_board_[getWest(getSouth(pos))],
                      go_board_[getEast(pos)], go_board_[pos], go_board_[getWest(pos)],
                      go_board_[getEast(getNorth(pos))], go_board_[getNorth(pos)], go_board_[getWest(getNorth(pos))]};

    pattern_go_board_[pos] = 0;
    for (int i = 0; i < 9; ++i) {
        pattern_go_board_[pos] <<= 2;
        if (pattern[i] == White) {
            ++pattern_go_board_[pos];
        }
        else if (pattern[i] == Empty) {
            pattern_go_board_[pos] += 2;
        }
        else if (pattern[i] == Wall) {
            pattern_go_board_[pos] += 3;
        }
    }
}

bool GoBoard::isNick(int pos, int color)
{
    if (go_board_[pos] != Empty) {
        return false;
    }
    int around_pos[4] = {getNorthWest(pos), getNorthEast(pos), getSouthWest(pos), getSouthEast(pos)};
    int inverse_colro = (color == Black ? White : Black);

    if (pos == 28) {
        if (go_board_[42] == inverse_colro) {
            return true;
        }
        else {
            return false;
        }
    }
    else if (pos == 36) {
        if (go_board_[48] == inverse_colro) {
            return true;
        }
        else {
            return false;
        }
    }
    else if (pos == 132) {
        if (go_board_[120] == inverse_colro) {
            return true;
        }
        else {
            return false;
        }
    }
    else if (pos == 140) {
        if (go_board_[126] == inverse_colro) {
            return true;
        }
        else {
            return false;
        }
    }
    else if (go_board_[around_pos[0]] == inverse_colro) {
        if (go_board_[around_pos[1]] == inverse_colro
                || go_board_[around_pos[2]] == inverse_colro
                || go_board_[around_pos[3]] == inverse_colro) {
            return true;
        }
        return false;
    }
    else if (go_board_[around_pos[1]] == inverse_colro) {
        if (go_board_[around_pos[0]] == inverse_colro
                || go_board_[around_pos[2]] == inverse_colro
                || go_board_[around_pos[3]] == inverse_colro) {
            return true;
        }
        return false;
    }
    else if (go_board_[around_pos[2]] == inverse_colro) {
        if (go_board_[around_pos[0]] == inverse_colro
                || go_board_[around_pos[1]] == inverse_colro
                || go_board_[around_pos[3]] == inverse_colro) {
            return true;
        }
        return false;
    }
    else if (go_board_[around_pos[3]] == inverse_colro) {
        if (go_board_[around_pos[0]] == inverse_colro
                || go_board_[around_pos[1]] == inverse_colro
                || go_board_[around_pos[2]] == inverse_colro) {
            return true;
        }
        return false;
    }

    return false;
}

void GoBoard::printProbability(std::array<double,  karray_board_size> board)
{
    std::cout << "= ";
    double r = 0, g = 0, b = 0;
    int count = 0;

    for (int i = kboard_start_index + 2; i < kboard_end_index; ++i) {

        // int index = i * kexpanded_board_size + j + kexpanded_board_size + 1;
        if (go_board_[i] == Black || go_board_[i] == White) {
                std::cout << "\"\" ";
        }
        else if (go_board_[i] == Empty){
            if (board[i] > 1) {
                board[i] = 1.0;
            }
            r = board[i] * 255.0;
            b = 255 - r;

            std::cout.fill('0');
            std::cout << "#" << std::hex << std::setw(2) << (int)r << std::setw(2)<< (int)g << std::setw(2) << (int)b << ' ';
        }
        if (!(i % kexpanded_board_size)) {
            std::cout << std::endl;
        }
    }
    std::cout << std::endl << std::endl;
}

void GoBoard::printProbabilityCount(std::array<double, karray_board_size> board)
{
    std::cout << "= ";
    int count = 0;

    for (int i = kboard_start_index + 2; i < kboard_end_index; ++i) {

        // int index = i * kexpanded_board_size + j + kexpanded_board_size + 1;
        if (go_board_[i] == Black || go_board_[i] == White) {
                std::cout << "\"\" ";
        }
        else if (go_board_[i] == Empty){
            std::cout << board[i] << " ";
        }
        if (!(i % kexpanded_board_size)) {
            std::cout << std::endl;
        }
    }
    std::cout << std::endl << std::endl;
}

void GoBoard::clearGoBoard()
{
    go_board_.fill(2);
    for (int i = 0, shift =  karray_board_size - 2 * kexpanded_board_size; i < kboard_start_index; ++i) {
        go_board_[i] = 3;
        go_board_[shift + i] = 3;
    }

    for (int i = kexpanded_board_size; i <=  karray_board_size - kexpanded_board_size; i += kexpanded_board_size) {
        go_board_[i - 1] = go_board_[i - 2] = go_board_[i] = go_board_[i + 1] = 3;
    }

    setPatternGoBoard();
}

void GoBoard::clearGoStrings()
{
    go_string_index_[0] = 1;
    go_string_index_[1] = 2;
    go_string_board_.fill(0);
    go_string_liberty_count_.resize(2 *  karray_board_size, 0);
}

void GoBoard::setBitPattern(int pos, int color, int bit_pos)
{
    if (color == Black) {
        pattern_go_board_[pos] = pattern_go_board_[pos] & ~(1 << (2 * bit_pos));
        pattern_go_board_[pos] = pattern_go_board_[pos] & ~(1 << (2 * bit_pos + 1));
    }
    else if (color == White) {
        pattern_go_board_[pos] = pattern_go_board_[pos] | (1 << (2 * bit_pos));
        pattern_go_board_[pos] = pattern_go_board_[pos] & ~(1 << (2 * bit_pos + 1));

    }
    else if (color == Empty) {
        pattern_go_board_[pos] = pattern_go_board_[pos] & ~(1 << (2 * bit_pos));
        pattern_go_board_[pos] = pattern_go_board_[pos] | (1 << (2 * bit_pos + 1));
    }
    else if (color == Wall) {
        pattern_go_board_[pos] = pattern_go_board_[pos] | (1 << (2 * bit_pos));
        pattern_go_board_[pos] = pattern_go_board_[pos] | (1 << (2 * bit_pos + 1));
    }
}



