#include <iostream>
#include <sstream>
#include <iomanip>
#include <array>
#include <fstream>
#include <memory>
#include <array>
#include <chrono>
#include <unordered_map>
#include <iterator>
#include "gtp.h"
#include "go_board.h"
#include "global.h"
#include "playout.h"

extern MakePattern make_pattern;

GTP::GTP()
{
    ai_name_ = "hikaru";
    handler_["name"] = &GTP::name;
    handler_["quit"] = &GTP::quit;
    handler_["play"] = &GTP::play;
    handler_["clear_board"] = &GTP::clearBoard;
    handler_["genmove"] = &GTP::genmove;
    handler_["boardsize"] = &GTP::boardSize;
    handler_["list_commands"] = &GTP::listCommands;
    handler_["gogui_analyze_commands"] = &GTP::goguiAnalyzeCommands;
    handler_["probability-analyst"] = &GTP::ProbabilityAnalyst;
    handler_["probability-pattern"] = &GTP::ProbabilityPattern;
    handler_["test_search"] = &GTP::doTestSearch;
    handler_["probability-pattern_count"] = &GTP::ProbabilityCountPattern;
    handler_["final_score"] = &GTP::finalScore;
    turn_color_ = Black;

    // move_frequency_.doPreparation();

    // make_pattern.testMakePattern();

    turn_count_ = 0;
    search_.clearNodes();

}

void GTP::loop() {
    std::cin.clear();
    std::cout.flush();

    Status status = OK;
    // std::cin.ignore();
    // std::cin.clear();

    while (status == OK) {
        std::string str;
        std::getline(std::cin, str);


        if (str.size() <= 1|| str[0] == '¥0') {
            std::cerr << "exit" << std::endl;
            return;
        }

        std::stringstream ss(str);
        std::string cmd, arg;
        ss >> cmd;
        std::getline(ss, arg);

        size_t first = arg.find_first_not_of(" ");
        size_t last = arg.find_last_not_of(" ");
        if (last != std::string::npos) {
            arg = arg.substr(first, last);
        }
        else {
            arg = "";
        }
        std::unordered_map<std::string, func>::const_iterator it = handler_.find(cmd);

        if (it != handler_.end()) {
            status = (this->*(it->second))(arg);
            std::cout << std::endl << std::endl;
        }
        else {
            failure("unknown command");
        }
    }
}

GTP::Status GTP::success(const std::string &arg)
{
    std::cout << "= " << arg << std::endl << std::endl;
    return OK;
}

GTP::Status GTP::failure(const std::string &arg)
{
    std::cout << "? " << arg << std::endl << std::endl;
    return OK;
}

GTP::Status GTP::name(const std::string &arg)
{
    return success(ai_name_);
}

GTP::Status GTP::quit(const std::string &arg)
{
    success("");
    return QUIT;
}

GTP::Status GTP::genmove(const std::string &arg)
{
    int color = (arg[0] == 'b' ? Black : White);
    search_.setGoBoard(go_board_);


    auto from = std::chrono::system_clock::now();

    int pos = search_.doCEDECSearch(color, turn_count_);

    // std::cerr << std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::system_clock::now() - from).count() << std::endl;

    if (pos == 0) {
        std::cout << "= PASS" << std::endl << std::endl;
            go_board_.initializeKoPoint();
        }
    else if (pos == -1) {
        std::cout << "= resign" << std::endl << std::endl;
    }
    else {
        int x = (pos % kexpanded_board_size) - 2;
        if (x >= 8) {
            ++x;
        }

        int y = pos / kexpanded_board_size - 2;
        // std::cerr << pos << " " << x << " " << y  << std::endl;
        std::cout << "= " << (char)('A' + x) << kboard_size - y << std::endl << std::endl;
        go_board_.setStone(pos, color);
        ++turn_count_;
        turn_color_ = (color == Black ? White : Black);
        // go_board_.printGoBoard();
        // go_board_.printGoStrings();
    }
    search_.clearNodes();

    return OK;
}

GTP::Status GTP::clearBoard(const std::string &arg)
{
    turn_count_ = 0;
    go_board_.initializeGoBoard();
    return success("");
}

GTP::Status GTP::boardSize(const std::string &arg)
{
    return success("");
}

GTP::Status GTP::listCommands(const std::string &arg)
{
    std::cout << "=0" << std::endl;
    std::cout << "test_search" << std::endl;
    std::cout << "gogui_analyze_commands" << std::endl << std::endl;
    return OK;
}

GTP::Status GTP::play(const std::string &arg)
{
    std::stringstream ss(arg.substr(3));

    if (arg[3] == 'P') {
        go_board_.setStone(0, (arg[0] == 'B' ? Black : White));
    }
    else {

        int x, y;
        ss >> y;
        x = arg[2] - 'A' + 1;
        if (arg[2] >= 'J') {
            --x;
        }
        y = kboard_size - y + 1;

        // std::cerr << x << " " << y << std::endl;
        int pos = x + y * kexpanded_board_size + 1 + kexpanded_board_size;

        // std::cerr << pos << std::endl;
        go_board_.setStone(pos, (arg[0] == 'B' ? Black : White));
    }
    turn_color_ = (arg[0] == 'B' ? White : Black);
    ++turn_count_;
    // go_board_.printGoStrings();

    return success("");
}

GTP::Status GTP::goguiAnalyzeCommands(const std::string &arg)
{
    std::cout << "= cboard/Proba/probability-analyst" << std::endl;
    std::cout << " cboard/LegalAnd/probability-pattern l" << std::endl;
    std::cout << " cboard/Pattern3x3/probability-pattern 3x3" << std::endl;
    std::cout << " cboard/Pattern5x5/probability-pattern 5x5" << std::endl;
    std::cout << " cboard/Pattern9x9/probability-pattern 9x9" << std::endl;
    std::cout << " cboard/Pattern3x3WithPosition/probability-pattern w3x3" << std::endl;
    std::cout << " sboard/PatternCount3x3/probability-pattern_count 3x3" << std::endl;
    std::cout << " sboard/PatternCount5x5/probability-pattern_count 5x5" << std::endl;
    std::cout << " sboard/PatternCount3x3WitPosition/probability-pattern_count w3x3" << std::endl;
    std::cout << " sboard/Translation/probability-pattern_count t2" << std::endl << std::endl;
    return OK;
}

GTP::Status GTP::ProbabilityAnalyst(const std::string &arg)
{
    go_board_.printProbability(move_frequency_.getFrequencyBoard(go_board_.getGoBoard(), turn_color_));
    return OK;
}

GTP::Status GTP::ProbabilityPattern3x3(const std::string &)
{
    go_board_.printProbability(search_.getFrequencyBoard(go_board_, turn_color_, 0));
    return OK;
}

GTP::Status GTP::ProbabilityCountPattern(const std::string &arg)
{
    go_board_.setPatternGoBoard();
    if (arg[0] == '3') {
       go_board_.printProbabilityCount(make_pattern.getFrequencyBoard(go_board_, turn_color_, 0));
    }
    else if (arg[0] == '5') {
       go_board_.printProbabilityCount(make_pattern.getFrequencyBoard(go_board_, turn_color_, 1));
    }
    else if (arg[0] == '9') {
       go_board_.printProbabilityCount(make_pattern.getFrequencyBoard(go_board_, turn_color_, 2));
    }
    else if (arg[0] == 'w') {
       go_board_.printProbabilityCount(make_pattern.getFrequencyBoard(go_board_, turn_color_, 3));
    }
    else if (arg[0] == 't') {
       go_board_.printProbabilityCount(make_pattern.getFrequencyBoard(go_board_, turn_color_, 4));
    }
    return OK;
}

GTP::Status GTP::ProbabilityPattern5x5(const std::string &)
{
    go_board_.printProbability(search_.getFrequencyBoard(go_board_, turn_color_, 1));
    return OK;
}

GTP::Status GTP::ProbabilityPattern9x9(const std::string &)
{
    go_board_.printProbability(search_.getFrequencyBoard(go_board_, turn_color_, 2));
    return OK;
}

GTP::Status GTP::ProbabilityPattern3x3WithPosition(const std::string &)
{
    go_board_.printProbability(search_.getFrequencyBoard(go_board_, turn_color_, 3));
    return OK;
}

GTP::Status GTP::ProbabilityPattern(const std::string &arg)
{
    go_board_.setPatternGoBoard();
    if (arg[0] == '3') {
       go_board_.printProbability(make_pattern.getFrequencyBoard(go_board_, turn_color_, 0));
    }
    else if (arg[0] == '5') {
       go_board_.printProbability(make_pattern.getFrequencyBoard(go_board_, turn_color_, 1));
    }
    else if (arg[0] == '9') {
       go_board_.printProbability(make_pattern.getFrequencyBoard(go_board_, turn_color_, 2));
    }
    else if (arg[0] == 'w') {
       go_board_.printProbability(make_pattern.getFrequencyBoard(go_board_, turn_color_, 3));
    }
    else if (arg[0] == 't') {
       go_board_.printProbability(make_pattern.getFrequencyBoard(go_board_, turn_color_, 4));
    }
    else if (arg[0] == 'l') {
        std::array<double, karray_board_size> legal_board;
        legal_board.fill(0.0);
        for (int i = 0; i < kboard_square_size; ++i) {
            int move = board9x9_index[i];
            if ((go_board_.isNick(move, turn_color_) == true || go_board_.isEye(move, turn_color_) == false)
                    && go_board_.isLegalMove(move, turn_color_) == true) {
                legal_board[move] = 1;
            }
        }
        go_board_.printProbability(legal_board);
    }
    return OK;
}

GTP::Status GTP::doTestSearch(const std::string &)
{

    make_pattern.printTransitionCount(go_board_, turn_color_);


    return success("");
}

GTP::Status GTP::finalScore(const std::string &)
{
    search_.setGoBoard(go_board_);
    std::cout << "= " << search_.getResultChinese(turn_color_) << std::endl << std::endl;

    return OK;
}
