#ifndef MAKE_PATTERN_H
#define MAKE_PATTERN_H

#include <bitset>
#include <vector>
#include <array>
#include <unordered_map>
#include <iostream>
// #include <google/sparse_hash_map>
#include <queue>
#include "global.h"
#include "go_board.h"

class MakePattern
{
public:
    MakePattern();
    ~MakePattern() { std::cerr << "pattern delete" << std::endl; }
    int getPattern3x3ToBit(const std::array<int, 9> &pattern);
    std::array<int, 9> getRotatePattern3x3(const std::array<int, 9> &pattern);
    std::array<int, 9> getReversePattern3x3(const std::array<int, 9> &pattern);

    void printPattern3x3(const std::array<int, 9> &pattern);
    void printPattern5x5(const std::array<int, 25> &pattern);
    void printBitPattern5x5(const std::bitset<50> &pattern);
    void printBitPattern9x9(const std::bitset<338> &pattern);
    void printPattern9x9(const std::array<int, karray_board_size> &pattern);
    void printPattern243(const std::bitset<243> &pattern);

    void testMakePattern();
    int getPattern9x9Count(const std::array<int, karray_board_size> &go_board, int pos, int color);
    std::array<std::pair<int, int>, kboard_square_size> getPatternCountBoard(const GoBoard &current_go_board, int color, int mode = 0);
    std::array<int, 9> getPattern3x3(const std::array<int, karray_board_size> go_board, int pos);

    void printTransitionCount(GoBoard go_board, int color);

    int getMinPattern3x3ToBit(std::array<int, 9> pattern);
    // void setColorToBitset50(std::bitset<50> &pattern, int color, int index);
    std::array<int, 25> getPattern5x5(const std::array<int, karray_board_size> &go_board, int pos);
    std::array<int, 25> getRotatePattern5x5(const std::array<int, 25> &pattern);
    std::array<int, 25> getReversePattern5x5(const std::array<int, 25> &pattern);
    std::bitset<50> getBitPattern5x5(const std::array<int, 25> &pattern);
    std::bitset<338> getBitPattern9x9(const std::array<int, karray_board_size> &pattern);

    std::array<int, karray_board_size> getRotatePattern9x9(const std::array<int, karray_board_size> &pattern);
    std::array<int, karray_board_size> getReversePattern9x9(const std::array<int, karray_board_size> &pattern);

    std::array<int, 9>  getBitToPattern3x3(int bit_pattern);
    void doMakePattern();
    std::vector<int> getSequentialMove(int pre_pos, int second_pre_pos, int third_pre_pos) { return four_transition_move_[third_pre_pos][second_pre_pos][pre_pos]; }
    int getResultBoard9x9FromPattern(const std::array<int, karray_board_size> &pattern_go_board);


    std::array<std::unordered_map<std::bitset<243>, std::array<int, 4> >, kboard_square_size> pattern243_count_;
    std::bitset<243> getPattern243(const std::array<int, karray_board_size> &go_board);
    std::bitset<243> getRotatePattern243(const std::bitset<243> &pattern);
    std::bitset<243> getReversePattern243(const std::bitset<243> &pattern);

    std::array<std::array<int, kboard_square_size>, krandom_table_size> random_table_;
    std::array<std::array<std::array<std::pair<int, int>, karray_board_size>, karray_board_size>, karray_board_size> transition_count_move_;


private:

    std::array<std::array<int, 262144>, 2> pattern3x3_count_;
    std::unordered_map<std::bitset<18>, int> black_bit_pattern3x3_count_;
    std::unordered_map<std::bitset<18>, int> white_bit_pattern3x3_count_;
    std::array<std::array<std::array<std::array<int, 262144>, karray_board_size>, 5>, 2> pattern3x3_with_pos_count_;
    std::array<int, karray_board_size> each_with_pos_move_count_;

    // [color][second_pre][pre][post]
    std::array<std::array<std::array<int, karray_board_size>,karray_board_size>, karray_board_size> transition_count_;
    std::array<std::array<int, karray_board_size>,karray_board_size> two_transition_count_;

    std::unordered_map<int, std::unordered_map<int, std::unordered_map<int, std::unordered_map<int, int> > > > four_transition_count_;
    std::unordered_map<int, std::unordered_map<int, std::unordered_map<int, std::vector<int> > > > four_transition_move_;


    std::unordered_map<std::bitset<50>, int> black_pattern5x5_count_;
    std::unordered_map<std::bitset<50>, int> white_pattern5x5_count_;
    std::unordered_map<std::bitset<338>, std::unordered_map<int, int> > black_pattern9x9_count_;
    std::unordered_map<std::bitset<338>, std::unordered_map<int, int> > white_pattern9x9_count_;
    std::unordered_map<std::array<int, 25>, int> pattern5x5_count_;

    std::array<int, 262144> result_pattern_;

    int getWallCount(std::array<int, 9> pattern3x3);

public:
    // assume black = 0, white = 1
    long getPattern3x3FrequencyCount(int bit_pattern, int color) { return pattern3x3_count_[color][bit_pattern]; }
    std::array<double,  karray_board_size> getFrequencyBoard(const GoBoard &go_board, int color, int mode);
};

#endif // MAKE_PATTERN_H
