#include <chrono>
#include <iostream>
#include <memory>
#include <random>
#include <algorithm>
#include <vector>
#include <queue>
#include "go_board.h"
#include "playout.h"
#include "global.h"
#include "make_pattern.h"


PlayOut::PlayOut()
{
    // initialization of random
    std::random_device rnd;

    // vector for initializatioon
    std::vector<uint_least32_t> v(10);
    generate(v.begin(), v.end(), ref(rnd));

    std::seed_seq seed(v.begin(), v.end());
    engine_ = std::make_shared<std::mt19937>(seed);
}

int PlayOut::doOnePlayOut(int color)
{
    int move_count = 0;
    int pass_count = 0;
    int pos = 0;

    for (; pass_count - 2;) {
        if(++move_count >  karray_board_size) {
            break;

        }

        pos = doRandomSearch(color);
        if (pos == 0) {
            ++pass_count;
        }
        else {
            playout_go_board_.setStone(pos, color);
            pass_count = 0;
        }
        color = (color == Black ? White : Black);
    }

    return playout_go_board_.getResultChineseRule(color);
}

int PlayOut::doRandomSearch(int color)
{
    std::array<int,  karray_board_size> legal_move;
    int legal_move_size = 0;
    int point_color;
    int pos;
    for (int i = 0; i <  kboard_square_size; ++i) {
        pos = board9x9_index[i];
        int around_pos[4] = {getEast(pos), getWest(pos), getNorth(pos), getSouth(pos)};

        if (((point_color = playout_go_board_.getPointColor(around_pos[0])) == color || point_color == Wall)
                && ((point_color = playout_go_board_.getPointColor(around_pos[1])) == color || point_color == Wall)
                && ((point_color = playout_go_board_.getPointColor(around_pos[2])) == color || point_color == Wall)
                && ((point_color = playout_go_board_.getPointColor(around_pos[3])) == color || point_color == Wall)) {

                if (playout_go_board_.getLibertyCount(around_pos[0]) > 1 && playout_go_board_.getLibertyCount(around_pos[1]) > 1
                        && playout_go_board_.getLibertyCount(around_pos[2]) > 1 && playout_go_board_.getLibertyCount(around_pos[3]) > 1) {
                    continue;
                }
        }

        // std::cerr << pos << " " << color << std::endl;
        if (playout_go_board_.isLegalMove(pos, color) == true) {
            legal_move[legal_move_size] = pos;
            ++legal_move_size;
        }
    }

    if (legal_move_size == 0) {
        return 0;
    }

    distribution_ = std::make_shared<std::uniform_int_distribution<int> >(0, legal_move_size - 1);

    return legal_move[(*distribution_)(*engine_)];
}


double PlayOut::operator ()(const GoBoard &go_board, int color)
{
    int win_count = 0;
    int total_count = 0;
    auto to = std::chrono::system_clock::now() + std::chrono::milliseconds(10);
    while (std::chrono::system_clock::now() < to) {
        playout_go_board_ = go_board;
        doOnePlayOut(color);
        win_count += playout_go_board_.getResultChineseRule(color);
        ++total_count;
    }

    return (double)win_count / (double)total_count;
}

