#include <random>
#include <vector>
#include <array>
#include <algorithm>
#include <sstream>
#include <iostream>
#include <chrono>
#include <thread>
#include <mutex>
#include <future>
#include <atomic>
#include <atomic>
#include <bitset>
#include "search.h"
#include "ucb_node.h"
#include "global.h"
#include "playout.h"
#include "uct_thread.h"
#include "make_pattern.h"

extern MakePattern make_pattern;

Search::Search()
{
    // initialization of random
    std::random_device rnd;

    // vector for initializatioon
    std::vector<uint_least32_t> v(10);
    generate(v.begin(), v.end(), ref(rnd));

    std::seed_seq seed(v.begin(), v.end());
    engine_ = std::make_shared<std::mt19937>(seed);

    thread_number_ = 1;
    thread_win_ratio_ = 0.0;
    search_depth_ = 1;
}

int Search::doRandomSearch(int color)
{
    std::array<int,  karray_board_size> legal_move;
    int legal_move_size = 0;

    for (int i = kboard_start_index; i < kboard_end_index; ++i) {
        if (work_go_board_.isEye(i, color)) {
            continue;
        }
        else if (work_go_board_.isLegalMove(i, color) == true) {
            legal_move[legal_move_size] = i;
            ++legal_move_size;
        }
    }

    if (legal_move_size == 0) {
        return 0;
    }

    distribution_ = std::make_shared<std::uniform_int_distribution<int> >(0, legal_move_size - 1);

    return legal_move[(*distribution_)(*engine_)];
}

void Search::doRandomPlayOut(int color)
{
    int move_count = 0;
    int pass_count = 0;
    int pos;
    for (; pass_count - 2;) {
        if(++move_count >  karray_board_size) {
            break;

        }

        pos = doRandomSearch(color);
        if (pos == 0) {
            ++pass_count;
        }
        else {
            work_go_board_.setStone(pos, color);
            pass_count = 0;
        }
        color = (color == Black ? White : Black);
    }

    // go_board_.printGoBoard();
}

int Search::doThreadOnePlayOut(GoBoard go_board, int color)
{
    int move_count = 0;
    int pass_count = 0;
    int pos = 0;

    for (; pass_count - 2;) {
        if(++move_count >  karray_board_size) {
            break;

        }

        pos = doThreadRandomSearch(go_board, color);
        if (pos == 0) {
            ++pass_count;
        }
        else {
            go_board.setStone(pos, color);
            pass_count = 0;
        }
        color = (color == Black ? White : Black);
    }

    return go_board.getResultChineseRule(color);
}

int Search::doThreadRandomSearch(const GoBoard &go_board, int color)
{
    // legal_move.reserve( karray_board_size);
    std::array<int,  karray_board_size> legal_move;
    int legal_move_size = 0;
    int point_color;
    int pos;
    for (int i = 0; i <  kboard_square_size; ++i) {
        pos = board9x9_index[i];
        int around_pos[4] = {getEast(pos), getWest(pos), getNorth(pos), getSouth(pos)};

        if (((point_color = go_board.getPointColor(around_pos[0])) == color || point_color == Wall)
                && ((point_color = go_board.getPointColor(around_pos[1])) == color || point_color == Wall)
                && ((point_color = go_board.getPointColor(around_pos[2])) == color || point_color == Wall)
                && ((point_color = go_board.getPointColor(around_pos[3])) == color || point_color == Wall)) {

                if (go_board.getLibertyCount(around_pos[0]) > 1 && go_board.getLibertyCount(around_pos[1]) > 1
                        && go_board.getLibertyCount(around_pos[2]) > 1 &&go_board.getLibertyCount(around_pos[3]) > 1) {
                    continue;
                }
        }

        // std::cerr << pos << " " << color << std::endl;
        if (go_board.isLegalMove(pos, color) == true) {
            // legal_move.push_back(i);
            legal_move[legal_move_size] = pos;
            ++legal_move_size;
        }
    }

    if (legal_move_size == 0) {
        return 0;
    }

    distribution_ = std::make_shared<std::uniform_int_distribution<int> >(0, legal_move_size - 1);

    return legal_move[(*distribution_)(*engine_)];
}

int Search::doUCBSearch(int color)
{
    std::cerr << "do ucb search" << std::endl;

    return 0;
}

void Search::testRandomPlayOut()
{
    for (int i = 0; i < 1; ++i) {
        std::thread t((Search()));
        t.detach();
    }
}

int Search::doCEDECSearch(int color, int turn)
{
    if (turn > 300) {
        return 0;
    }
/*
    if (turn == 0) {
        return 112;
        return 97;
    }
*/
    if (turn > 10 && current_go_board_.getPrePosition() == 0) {
        return 0;
    }


    std::array<double,  karray_board_size> freqency_board = make_pattern.getFrequencyBoard(current_go_board_, color, 4);

    std::unordered_map<int, double> bias;
    bias[0] = 0.0;

    int max_count = 0, max_count_pos = 0, tmp_count, pre_pos = 0, pre_max_count = 0, max_select_count = 0;
    int inverse_color = (color == Black ? White : Black);


    if (turn < 25) {
        for (int i = kboard_start_index; i < kboard_end_index; ++i) {
             if ((tmp_count = make_pattern.getPattern9x9Count(current_go_board_.getGoBoard(), i, color)) > 0) {
                 if (current_go_board_.isLegalMove(i, color) == true) {
                     int sum_count = 0;
                     int select_count = 0;
                     for (int j = 0;  j < kboard_square_size; ++j) {
                        work_go_board_ = current_go_board_;
                        work_go_board_.setStone(i, color);
                        if (i == board9x9_index[j]) {
                            continue;
                        }
                        work_go_board_.setStone(board9x9_index[j], inverse_color);
                        for (int k = 0; k < kboard_square_size; ++k) {
                            if (board9x9_index[k] == i || board9x9_index[k] == board9x9_index[j]) {
                                continue;
                            }
                            select_count += make_pattern.getPattern9x9Count(work_go_board_.getGoBoard(), board9x9_index[k], color);
                            if (make_pattern.getPattern9x9Count(work_go_board_.getGoBoard(), board9x9_index[k], color)) {                                
#ifdef EBUG
                                std::cerr << getCharacterPosition(board9x9_index[j]) << " - " << getCharacterPosition(board9x9_index[k]) << " : "
                                            << make_pattern.getPattern9x9Count(work_go_board_.getGoBoard(), board9x9_index[k], color) << std::endl;

#endif
                            }
                            sum_count += (make_pattern.getPattern9x9Count(work_go_board_.getGoBoard(), board9x9_index[k], color) > 0 ? 1 : 0);
                        }
                     }
#ifdef EBUG
                     std::cerr << "total : " << sum_count << " " << select_count << std::endl;
#endif
                     if (sum_count > max_count) {
                         pre_pos = max_count_pos;
                         max_count = sum_count;
                         max_count_pos = i;
                         max_select_count = select_count;
                     }
                     else if (sum_count == max_count && select_count > max_select_count) {
                         pre_pos = max_count_pos;
                         max_count = sum_count;
                         max_count_pos = i;
                         max_select_count = select_count;
                     }
                 }
             }
        }
    }


    if (max_count_pos != 0) {
        return max_count_pos;
    }

    current_go_board_.setPatternGoBoard();

    std::array<std::pair<int, int>, kboard_square_size> pattern_count = make_pattern.getPatternCountBoard(current_go_board_, color, 2);


    // set nodes of dept 0
    for (int i = 0; i < kthread_size; ++i) {
        nodes_[i][0].parent_index_ = 0;
        nodes_[i][0].pos_ = -1;
        nodes_[i][0].trial_count_ = 0;
        nodes_[i][0].child_index_ = 1;
        nodes_[i][0].depth_ = 0;
    }

    std::array<int, kthread_size> depth_one_node_end ={};
    int thread_node_size = kboard_square_size / kthread_size;
    for (int i = 1, j = 0, m = 1, n = 0; i <= kboard_square_size; ++i) {
        nodes_[n][m].pos_ = pattern_count[j].second;
        nodes_[n][m].depth_ = 1;
        nodes_[n][m].root_pos_ = pattern_count[j].second;
        j += kthread_size;
        ++m;
        if (j >= kboard_square_size) {
            depth_one_node_end[n] = m + 1;  // is next begin < x   not <=
            nodes_[n][m].pos_ = 0;
            nodes_[n][m].depth_ = 1;
            nodes_[n][m].root_pos_ = 0;
            m = 1;
            j = i / thread_node_size;
            ++n;
        }
    }

    std::array<std::thread, kthread_size> threads;
    for (int i = 0; i < kthread_size; ++i) {
        threads[i] = std::thread((UCTThread()),current_go_board_, color, &nodes_[i], depth_one_node_end[i], i, turn);
    }

    std::for_each(threads.begin(), threads.end(), std::mem_fn(&std::thread::join));

    std::array<std::pair<double, int>, kdepth_one_node_size> moves;
    double max_win_ratio = -1.0 * karray_board_size; // 0.0;
    double tmp;
    max_count = 0;
    int select_move = 0;
    int select_move_index = 0;
    int select_thread_number = 0;
    int sum_trial_count = 0;


    for (int i = 0, k = 0; i < kthread_size; ++i) {
        for (int j = 1; j < depth_one_node_end[i]; ++j) {

            double max_grandchild_win_ratio = 0.0, grand_child_win_ratio, select_grandchild_index = 0;


            for (int k = nodes_[i][j].child_index_; k < nodes_[i][j].child_index_ + knode_max_width; ++k) {
                grand_child_win_ratio = (double)nodes_[i][k].win_count_ / (double)nodes_[i][k].trial_count_;
                if (grand_child_win_ratio > max_grandchild_win_ratio) {
                    max_grandchild_win_ratio = grand_child_win_ratio;
                    select_grandchild_index = k;
                }
            }



#ifdef EBUG

            if ((tmp = (double)nodes_[i][j].win_count_ / (double)nodes_[i][j].trial_count_ - max_grandchild_win_ratio) > max_win_ratio) {
                max_win_ratio = tmp;
                select_move = nodes_[i][j].pos_;
                select_move_index = j;
                select_thread_number = i;
            }



            if (max_grandchild_win_ratio < min_grandchild_win_ratio) {
                min_grandchild_win_ratio = max_grandchild_win_ratio;
            }        
#endif

            // freqency_board

            if ((tmp = (double)nodes_[i][j].win_count_ / (double)nodes_[i][j].trial_count_ + bias[nodes_[i][j].pos_]
                 /*+ freqency_board[nodes_[i][j].pos_]*/)> max_win_ratio) {
            // if ((tmp = (double)nodes_[i][j].score_ / (double)nodes_[i][j].trial_count_) > max_win_ratio) {
                max_win_ratio = tmp;
                select_move = nodes_[i][j].pos_;
                select_move_index = j;
                select_thread_number = i;
            }

            moves[k] = std::make_pair(tmp, nodes_[i][j].pos_);
#ifdef EBUG
            std::cerr << k << " : " << getCharacterPosition(nodes_[i][j].pos_)
                      << " " << nodes_[i][j].win_count_ << " " << nodes_[i][j].trial_count_ << " | "
                      <<  (double) nodes_[i][j].win_count_ / (double)nodes_[i][j].trial_count_
                      << " " << nodes_[i][j].score_ // (nodes_[i].trial_count_ == 0 ? 1 : nodes_[i].trial_count_)
                      << " " << nodes_[i][j].trial_count_ << " " << (double)nodes_[i][j].score_ / (double)nodes_[i][j].trial_count_
                      << " " << max_grandchild_win_ratio
                      << " " << nodes_[i][j].ucb_value_
                      << " " << nodes_[i][j].depth_ << std::endl;

#endif

            ++k;
        }

    }


    // pass
    if ((double)nodes_[0][depth_one_node_end[0] - 1].win_count_ / (double) nodes_[0][depth_one_node_end[0] - 1].trial_count_ > 0.90
            && (double)nodes_[0][depth_one_node_end[0] - 1].win_count_ / (double) nodes_[0][depth_one_node_end[0] - 1].trial_count_ + 0.3 >= max_win_ratio) {
        return 0;
    }

    int sum_root_count = 0;
    for (int i = 0; i < kthread_size; ++i) {
        sum_root_count += nodes_[i][0].trial_count_;
    }
    std::sort(moves.rbegin(), moves.rend());

    if (max_win_ratio < 0.2) {
        return -1;
    }


    if (max_win_ratio < 0.4) {
        // max score
        double tmp_score;
        double max_score = -1.0 * karray_board_size;
        for (int i = 0; i < kthread_size; ++i) {
            for (int j = 0; j < depth_one_node_end[i]; ++ j) {
                if (nodes_[i][j].trial_count_ > 0 && (tmp_score = (double)nodes_[i][j].score_ / (double)nodes_[i][j].trial_count_) > max_score
                        && nodes_[i][j].score_ != 0) {
                    max_score = tmp_score;
                    select_move = nodes_[i][j].pos_;
                    select_thread_number = i;
                    select_move_index = j;
                    std::cerr << "0.2 smile " << getCharacterPosition(nodes_[i][j].pos_) << std::endl;
                }
            }
        }
    }


    if (select_move == 0) {
        for (int i = 0; i < kdepth_one_node_size; ++i) {
            if ((current_go_board_.isNick(moves[i].second, color) == true || current_go_board_.isEye(moves[i].second, color) == false)
                    && current_go_board_.isLegalMove(moves[i].second, color) == true && moves[i].second != 0) {
                select_move = moves[i].second;
                break;
            }
        }
    }


    std::cerr << "sum trial count " << sum_trial_count << std::endl;
    std::cerr << "thread number " << select_thread_number << std::endl;
    std::cerr << "thread index " << select_move_index << std::endl;

    int child_node = nodes_[select_thread_number][select_move_index].child_index_; //nodes_[select_move_index].child_index_;

#ifdef EBUG
    std::cerr << "parent : " << nodes_[select_thread_number][child_node].parent_index_ << std::endl;
    for (int i = child_node; i < child_node + 82; ++i) {
        std::cerr << getCharacterPosition(nodes_[select_thread_number][i].pos_) << " " << nodes_[select_thread_number][i].trial_count_ << " "
                  << nodes_[select_thread_number][i].win_count_ << " "
                  << (double)nodes_[select_thread_number][i].win_count_ / (double)nodes_[select_thread_number][i].trial_count_ << " "
                  << nodes_[select_thread_number][i].getUCBValue(nodes_[select_thread_number][select_move_index].trial_count_)
                  << " " << nodes_[select_thread_number][i].ucb_value_ << std::endl;
    }
#endif

    double child_max_win_ratio;
    int child_index;
    for (int i  = 0; i < 10; ++i) {
        std::cerr << select_move_index << " " << (double)nodes_[select_thread_number][select_move_index].trial_count_ << " "
                  << getCharacterPosition(nodes_[select_thread_number][select_move_index].pos_) << " "
                  << (double)nodes_[select_thread_number][select_move_index].win_count_ / (double)nodes_[select_thread_number][select_move_index].trial_count_
                  << std::endl;
        child_max_win_ratio = -1.0;
        if ((child_index = nodes_[select_thread_number][select_move_index].child_index_) == 0 || nodes_[select_thread_number][select_move_index].trial_count_ == 1) {
            break;
        }
        for (int j = child_index; j < kboard_square_size + child_index; ++j) {
            double tmp = (double)nodes_[select_thread_number][j].win_count_ / (double)nodes_[select_thread_number][j].trial_count_;
            if (tmp > child_max_win_ratio) {
                child_max_win_ratio = tmp;
                select_move_index = j;
            }
        }
    }


    if (select_move > 0) {
        return select_move;
    }
    else {
        return 0;
    }

    return doRandomSearch(color);

    return 0;
}

int Search::doPatternSearch(int color, int turn_count)
{
    if (turn_count == 0) {
        return 84;
    }

    double max_pattern_value = -1, tmp;
    int max_pattern_value_move = 0;
    for (int i = 0; i < kboard_square_size; ++i) {
        work_go_board_ = current_go_board_;
        if (work_go_board_.isLegalMove(board9x9_index[i], color) == true && work_go_board_.isEye(board9x9_index[i], color) == false) {
            work_go_board_.setStone(board9x9_index[i], color);
            std::bitset<243> board = make_pattern.getPattern243(work_go_board_.getGoBoard());
            if (color == Black) {
                for (auto it = make_pattern.pattern243_count_[work_go_board_.getStoneCount()].begin(); it != make_pattern.pattern243_count_[work_go_board_.getStoneCount()].end(); ++it) {
                    if (it->second[1] != 0) {
                        tmp = (double)(it->first & board).count() / (double)work_go_board_.getStoneCount() * (double)it->second[0] / ((double)it->second[1] + 0.1);
                        if (tmp > max_pattern_value) {
                            max_pattern_value = tmp;
                            max_pattern_value_move = board9x9_index[i];
                        }
                    }
                }
            }
            else if (color == White) {
                for (auto it = make_pattern.pattern243_count_[work_go_board_.getStoneCount()].begin(); it != make_pattern.pattern243_count_[work_go_board_.getStoneCount()].end(); ++it) {
                    if (it->second[3] != 0) {
                        tmp = (double)(it->first & board).count() / (double)work_go_board_.getStoneCount() * (double)it->second[2] / ((double)it->second[3] + 0.1);
                        if (tmp > max_pattern_value) {
                            max_pattern_value = tmp;
                            max_pattern_value_move = board9x9_index[i];
                        }
                    }
                }
            }
        }
    }

    std::cerr << max_pattern_value << std::endl;

    return max_pattern_value_move;
}

int Search::doSequentialRandomPlayOut(int color, int turn)
{
    if (turn < 4) {
        return doRandomSearch(color);
    }


    int pre_pos = current_go_board_.getPrePosition();
    int second_pre_pos = current_go_board_.getSecondPrePosition();
    int third_pre_pos = current_go_board_.getThirdPrePosition();

    std::vector<int> moves = make_pattern.getSequentialMove(pre_pos, second_pre_pos, third_pre_pos);


    std::random_device rd;
    std::mt19937 engine(rd());
    std::shuffle(moves.begin(), moves.end(), engine);

    for (int i = 0; i < moves.size(); ++i) {
        if (current_go_board_.isEye(moves[i], color) == false && current_go_board_.isLegalMove(moves[i], color) == true) {
            return moves[i];
        }
    }

    return doRandomSearch(color);

}

void Search::doTest()
{
    int win_count = 0;
    // std::thread((PlayOut()), current_go_board_, Black, win_count).join();
    std::future<double> thread = std::async(std::launch::async, (PlayOut()), current_go_board_, Black);


    std::cerr << thread.get() << std::endl;
}

void Search::printPoint(int pos)
{
    std::cerr << (char)(pos % kexpanded_board_size - 2 + 'A') << (kexpanded_board_size - pos / kexpanded_board_size - 2);
}

void Search::doThreadPlayOut(const GoBoard &go_board, int color, int num)
{
    // int win_coutn = 0;

    GoBoard playout_go_board;
    for (int i = 0; i < max_thread_playout_; ++i) {
        playout_go_board = go_board;
        // std::unique_lock<std::mutex> lk(mut_);
        thread_win_ratio_ += doThreadOnePlayOut(playout_go_board, color);
        // ++thread_win_count_;
        // std::cout << i << " times " << num << std::endl;
    }



    // std::cerr << thread_win_count_ << " / " << max_thread_playout_ << std::endl;
}

void Search::operator ()()
{
    static int count = 0;
    ++count;
    GoBoard go_board = current_go_board_;
    auto to = std::chrono::system_clock::now() + std::chrono::milliseconds(1000);
    while (std::chrono::system_clock::now() < to) {
        setGoBoard(go_board);
        doRandomPlayOut(Black);
        ++count;
        work_go_board_.printGoBoard();
    }
    std::cerr << count << std::endl;
}

std::string Search::getResultChinese(int color)
{
    std::stringstream ss;
    std::string str;
    doRandomPlayOut(color);
    double score = work_go_board_.getResultScoreChineseRule();

    // work_go_board_.printGoBoard();

    if (score > 0) {
        ss << "B+" << score;
    }
    else {
        ss << "W+" << -1.0 * score;
    }
    ss >> str;
    return str;
}

void Search::clearNodes()
{
    for (long long int i = 0; i < knode_size; ++i) {
        for (int j = 0; j < kthread_size; ++j) {
            nodes_[j][i].win_count_ = 0;
            nodes_[j][i].pos_ = 0;
            nodes_[j][i].parent_index_ = 0;
            nodes_[j][i].trial_count_ = 0;
            nodes_[j][i].child_index_ = 0;
            nodes_[j][i].depth_ = 0;
            nodes_[j][i].score_ = 0;
            nodes_[j][i].ucb_value_ = -0.1;
            nodes_[j][i].is_legal_move_ = true;
        }
    }
}

void Search::doThreeLookAhead()
{

}

int Search::doAttachAndBlock(int color)
{
    int inverse_color = (color == Black ? White : Black);
    int depo_color[2] = {0};
    int pos;
    // diff sitahane

    // tome
    for (int i = 2; i < 7; ++i) {
        pos = board9x9_upper_side_index[i];
        if (current_go_board_.getPointColor(pos) == inverse_color
                && current_go_board_.getPointColor(getSouth(pos)) == color) {
            depo_color[0] = current_go_board_.getPointColor(getSouthWest(pos));
            depo_color[1] = current_go_board_.getPointColor(getSouthEast(pos));
            if (depo_color[0] == inverse_color && depo_color[1] != inverse_color) {
                if (current_go_board_.isLegalMove(pos + 1, color)) {
                    return pos + 1;
                }
            }
            else if (depo_color[0] != inverse_color && depo_color[1] == inverse_color) {
                if (current_go_board_.isLegalMove(pos - 1, color)) {
                    return pos - 1;
                }
            }
        }
        pos = board9x9_under_side_index[i];
        if (current_go_board_.getPointColor(pos) == inverse_color
                && current_go_board_.getPointColor(getNorth(pos)) == color) {
            depo_color[0] = current_go_board_.getPointColor(getNorthWest(pos));
            depo_color[1] = current_go_board_.getPointColor(getNorthEast(pos));
            if (depo_color[0] == inverse_color && depo_color[1] != inverse_color) {
                if (current_go_board_.isLegalMove(pos + 1, color)) {
                    return pos + 1;
                }
            }
            else if (depo_color[0] != inverse_color && depo_color[1] == inverse_color) {
                if (current_go_board_.isLegalMove(pos - 1, color)) {
                    return pos - 1;
                }
            }
        }

        pos = board9x9_left_side_index[i];
        if (current_go_board_.getPointColor(pos) == inverse_color
                && current_go_board_.getPointColor(getEast(pos)) == color) {
            depo_color[0] = current_go_board_.getPointColor(getNorthEast(pos));
            depo_color[1] = current_go_board_.getPointColor(getSouthEast(pos));
            if (depo_color[0] == inverse_color && depo_color[1] != inverse_color) {
                if (current_go_board_.isLegalMove(pos + kexpanded_board_size, color)) {
                    return pos + kexpanded_board_size;
                }
            }
            else if (depo_color[0] != inverse_color && depo_color[1] == inverse_color) {
                if (current_go_board_.isLegalMove(pos - kexpanded_board_size, color)) {
                    return pos - kexpanded_board_size;
                }
            }
        }

        pos = board9x9_right_side_index[i];
        if (current_go_board_.getPointColor(pos) == inverse_color
                && current_go_board_.getPointColor(getWest(pos)) == color) {
            depo_color[0] = current_go_board_.getPointColor(getNorthWest(pos));
            depo_color[1] = current_go_board_.getPointColor(getSouthWest(pos));
            if (depo_color[0] == inverse_color && depo_color[1] != inverse_color) {
                if (current_go_board_.isLegalMove(pos + kexpanded_board_size, color)) {
                    return pos + kexpanded_board_size;
                }
            }
            else if (depo_color[0] != inverse_color && depo_color[1] == inverse_color) {
                if (current_go_board_.isLegalMove(pos - kexpanded_board_size, color)) {
                    return pos - kexpanded_board_size;
                }
            }
        }
    }

    for (int i = 2; i < 7; ++i) {
        pos = board9x9_second_upper_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getNorth(pos)) == color) {
                if (current_go_board_.getPointColor(getNorthWest(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getNorthEast(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }
        }

        pos = board9x9_second_under_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getSouth(pos)) == color) {
                if (current_go_board_.getPointColor(getSouthWest(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getSouthEast(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }
        }

        pos = board9x9_second_left_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getWest(pos)) == color) {
                if (current_go_board_.getPointColor(getNorthWest(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getSouthWest(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }
        }

        pos = board9x9_second_right_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getEast(pos)) == color) {
                if (current_go_board_.getPointColor(getNorthEast(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getSouthEast(pos)) == inverse_color) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }
        }

    }

    for (int i = 2; i < 7; ++i) {
        pos = board9x9_second_upper_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getSouth(pos)) == color) {
                if (current_go_board_.getPointColor(getWest(pos)) == inverse_color
                        && current_go_board_.getPointColor(getNorthEast(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getEast(pos)) == inverse_color
                        && current_go_board_.getPointColor(getNorthWest(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }

        }


        pos = board9x9_second_under_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getNorth(pos)) == color) {
                if (current_go_board_.getPointColor(getWest(pos)) == inverse_color
                        && current_go_board_.getPointColor(getSouthEast(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getEast(pos)) == inverse_color
                        && current_go_board_.getPointColor(getSouthWest(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }

        }

        pos = board9x9_second_left_side_index[i];

        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getEast(pos)) == color) {
                if (current_go_board_.getPointColor(getNorth(pos)) == inverse_color
                        && current_go_board_.getPointColor(getSouthWest(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }

                if (current_go_board_.getPointColor(getSouth(pos)) == inverse_color
                        && current_go_board_.getPointColor(getNorthWest(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }

        }

        pos = board9x9_second_right_side_index[i];
        if (current_go_board_.getPointColor(pos) == Empty) {
            if (current_go_board_.getPointColor(getWest(pos)) == color) {
                if (current_go_board_.getPointColor(getNorth(pos)) == inverse_color
                        && current_go_board_.getPointColor(getSouthEast(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
                if (current_go_board_.getPointColor(getSouth(pos)) == inverse_color
                        && current_go_board_.getPointColor(getNorthEast(pos)) == Empty) {
                    if (current_go_board_.isLegalMove(pos, color)) {
                        return pos;
                    }
                }
            }

        }

    }

    return 0;
}

std::array<double, karray_board_size> Search::getFrequencyBoard(GoBoard go_board, int color, int mode)
{
    return make_pattern.getFrequencyBoard(go_board, color, mode);
}
