#ifndef SEARCH_H
#define SEARCH_H

#include <random>
#include <memory>
#include <iostream>
#include <string>
#include <list>
#include <atomic>
#include "go_board.h"
#include "global.h"
#include "make_pattern.h"
#include "ucb_node.h"

class Search
{
public:
    Search();
    Search(const Search&) {}

    int doRandomSearch(int color);
    void doRandomPlayOut(int color);
    void doThreadPlayOut(const GoBoard &go_board, int color, int num);
    int doThreadRandomSearch(const GoBoard &go_board, int color);
    int doThreadOnePlayOut(GoBoard go_board, int color);
    int doUCBSearch(int color);
    void testRandomPlayOut();
    int doCEDECSearch(int color, int turn = 0);
    int doPatternSearch(int color, int turn_count);
    int doSequentialRandomPlayOut(int color, int turn = 0);
    void printWorkBoard() { work_go_board_.printGoBoard(); }
    void doTest();
    void printPoint(int pos);
    void operator() ();
    std::string getResultChinese(int color);
    void clearNodes();

private:
    // static std::mutex mut_;

    GoBoard current_go_board_;
    GoBoard work_go_board_;

    void doThreeLookAhead();
    std::shared_ptr<std::mt19937> engine_;
    std::shared_ptr<std::uniform_int_distribution<int> > distribution_;
    int max_thread_playout_;
    double thread_win_ratio_;
    int thread_number_;
    int search_depth_;
    std::array<std::array<UCBNode, knode_size>, kthread_size> nodes_;
    std::list<int> sequential_moves_;

    int doAttachAndBlock(int color);
public:
    void setGoBoard(const GoBoard &go_board) {
                                        current_go_board_ = go_board;
                                        work_go_board_ = go_board;
                                      }
    std::array<double,  karray_board_size> getFrequencyBoard(GoBoard go_board, int color, int mode);
};

#endif // SEARCH_H
