#ifndef UCT_THREAD_H
#define UCT_THREAD_H

#include <array>
#include <atomic>
#include <random>
#include <memory>
#include <algorithm>
#include "ucb_node.h"
#include "global.h"
#include "go_board.h"

class UCTThread
{
public:
    UCTThread();
    void operator() (const GoBoard &go_board,const int color, std::array<UCBNode, knode_size> *nodes,
                     int depth_one_node_end, int thread_number, int turn);
private:
    std::array<int, kboard_square_size> random_index_;
    std::array<UCBNode, knode_size> *nodes_;
    int use_node_max_index_;
    GoBoard go_board_;
    int color_;
    int depth_;
    int thread_number_;
    int depth_one_node_end_;
    int turn_;

    std::mt19937 engine_;
    std::uniform_int_distribution<int> distribution_;
    std::uniform_int_distribution<long long int> random_table_distribution_;
    std::uniform_int_distribution<int> pibot_distribution_;
    std::normal_distribution<> normal_distributuion_;

    long long int getMaxUCBValueIndex(long long int parent_index, long long int start_index, long long range);  // 1 : pass
    void doOnePlayOut();
    void updateNode(int node_index, int result);
    void makeChildNode(int node_index, int color);

};

#endif // UCT_THREAD_H
